@@ -49,6 +49,18 @@ extern "C" {
 #  include "config.h"
 # endif
 
+# if HAVE_CLOCK_SYSCALL
+#  ifndef EV_USE_CLOCK_SYSCALL
+#   define EV_USE_CLOCK_SYSCALL 1
+#   ifndef EV_USE_REALTIME
+#    define EV_USE_REALTIME  0
+#   endif
+#   ifndef EV_USE_MONOTONIC
+#    define EV_USE_MONOTONIC 1
+#   endif
+#  endif
+# endif
+
 # if HAVE_CLOCK_GETTIME
 #  ifndef EV_USE_MONOTONIC
 #   define EV_USE_MONOTONIC 1
@@ -166,6 +178,14 @@ extern "C" {
 
 /* this block tries to deduce configuration from header-defined symbols and defaults */
 
+#ifndef EV_USE_CLOCK_SYSCALL
+# if __linux && __GLIBC__ >= 2
+#  define EV_USE_CLOCK_SYSCALL 1
+# else
+#  define EV_USE_CLOCK_SYSCALL 0
+# endif
+#endif
+
 #ifndef EV_USE_MONOTONIC
 # if defined (_POSIX_MONOTONIC_CLOCK) && _POSIX_MONOTONIC_CLOCK >= 0
 #  define EV_USE_MONOTONIC 1
@@ -289,6 +309,7 @@ extern "C" {
 
 #if EV_USE_INOTIFY
 # include <sys/utsname.h>
+# include <sys/statfs.h>
 # include <sys/inotify.h>
 /* some very old inotify.h headers don't have IN_DONT_FOLLOW */
 # ifndef IN_DONT_FOLLOW
@@ -301,6 +322,15 @@ extern "C" {
 # include <winsock.h>
 #endif
 
+/* on linux, we can use a (slow) syscall to avoid a dependency on pthread, */
+/* which makes programs even slower. might work on other unices, too. */
+#if EV_USE_CLOCK_SYSCALL
+# include <syscall.h>
+# define clock_gettime(id, ts) syscall (SYS_clock_gettime, (id), (ts))
+# undef EV_USE_MONOTONIC
+# define EV_USE_MONOTONIC 1
+#endif
+
 #if EV_USE_EVENTFD
 /* our minimum requirement is glibc 2.7 which has the stub, but not the header */
 # include <stdint.h>
@@ -1242,8 +1272,8 @@ ev_recommended_backends (void)
   flags &= ~EVBACKEND_KQUEUE;
 #endif
 #ifdef __APPLE__
-  // flags &= ~EVBACKEND_KQUEUE; for documentation
-  flags &= ~EVBACKEND_POLL;
+  // flags &= ~EVBACKEND_KQUEUE & ~EVBACKEND_POLL; for documentation
+  flags &= ~EVBACKEND_SELECT;
 #endif
 
   return flags;
@@ -1661,7 +1691,7 @@ ev_default_fork (void)
   struct ev_loop *loop = ev_default_loop_ptr;
 #endif
 
-  ev_loop_fork (EV_A);
+  postfork = 1; /* must be in line with ev_loop_fork */
 }
 
 /*****************************************************************************/
@@ -2416,8 +2446,9 @@ ev_child_stop (EV_P_ ev_child *w)
 #  define lstat(a,b) _stati64 (a,b)
 # endif
 
-#define DEF_STAT_INTERVAL 5.0074891
-#define MIN_STAT_INTERVAL 0.1074891
+#define DEF_STAT_INTERVAL  5.0074891
+#define NFS_STAT_INTERVAL 30.1074891 /* for filesystems potentially failing inotify */
+#define MIN_STAT_INTERVAL  0.1074891
 
 static void noinline stat_timer_cb (EV_P_ ev_timer *w_, int revents);
 
@@ -2431,10 +2462,11 @@ infy_add (EV_P_ ev_stat *w)
 
   if (w->wd < 0)
     {
-      ev_timer_start (EV_A_ &w->timer); /* this is not race-free, so we still need to recheck periodically */
+      w->timer.repeat = w->interval ? w->interval : DEF_STAT_INTERVAL;
+      ev_timer_again (EV_A_ &w->timer); /* this is not race-free, so we still need to recheck periodically */
 
       /* monitor some parent directory for speedup hints */
-      /* note that exceeding the hardcoded limit is not a correctness issue, */
+      /* note that exceeding the hardcoded path limit is not a correctness issue, */
       /* but an efficiency issue only */
       if ((errno == ENOENT || errno == EACCES) && strlen (w->path) < 4096)
         {
@@ -2448,8 +2480,8 @@ infy_add (EV_P_ ev_stat *w)
 
               char *pend = strrchr (path, '/');
 
-              if (!pend)
-                break; /* whoops, no '/', complain to your admin */
+              if (!pend || pend == path)
+                break;
 
               *pend = 0;
               w->wd = inotify_add_watch (fs_fd, path, mask);
@@ -2457,11 +2489,28 @@ infy_add (EV_P_ ev_stat *w)
           while (w->wd < 0 && (errno == ENOENT || errno == EACCES));
         }
     }
-  else
-    ev_timer_stop (EV_A_ &w->timer); /* we can watch this in a race-free way */
 
   if (w->wd >= 0)
-    wlist_add (&fs_hash [w->wd & (EV_INOTIFY_HASHSIZE - 1)].head, (WL)w);
+    {
+      wlist_add (&fs_hash [w->wd & (EV_INOTIFY_HASHSIZE - 1)].head, (WL)w);
+
+      /* now local changes will be tracked by inotify, but remote changes won't */
+      /* unless the filesystem it known to be local, we therefore still poll */
+      /* also do poll on <2.6.25, but with normal frequency */
+      struct statfs sfs;
+
+      if (fs_2625 && !statfs (w->path, &sfs))
+        if (sfs.f_type == 0x1373 /* devfs */
+            || sfs.f_type == 0xEF53 /* ext2/3 */
+            || sfs.f_type == 0x3153464a /* jfs */
+            || sfs.f_type == 0x52654973 /* reiser3 */
+            || sfs.f_type == 0x01021994 /* tempfs */
+            || sfs.f_type == 0x58465342 /* xfs */)
+          return;
+
+      w->timer.repeat = w->interval ? w->interval : fs_2625 ? NFS_STAT_INTERVAL : DEF_STAT_INTERVAL;
+      ev_timer_again (EV_A_ &w->timer);
+    }
 }
 
 static void noinline
@@ -2501,6 +2550,7 @@ infy_wd (EV_P_ int slot, int wd, struct inotify_event *ev)
             {
               if (ev->mask & (IN_IGNORED | IN_UNMOUNT | IN_DELETE_SELF))
                 {
+                  wlist_del (&fs_hash [slot & (EV_INOTIFY_HASHSIZE - 1)].head, (WL)w);
                   w->wd = -1;
                   infy_add (EV_A_ w); /* re-add, no matter what */
                 }
@@ -2523,32 +2573,38 @@ infy_cb (EV_P_ ev_io *w, int revents)
     infy_wd (EV_A_ ev->wd, ev->wd, ev);
 }
 
+void inline_size
+check_2625 (EV_P)
+{
+  /* kernels < 2.6.25 are borked
+   * http://www.ussg.indiana.edu/hypermail/linux/kernel/0711.3/1208.html
+   */
+  struct utsname buf;
+  int major, minor, micro;
+
+  if (uname (&buf))
+    return;
+
+  if (sscanf (buf.release, "%d.%d.%d", &major, &minor, &micro) != 3)
+    return;
+
+  if (major < 2
+      || (major == 2 && minor < 6)
+      || (major == 2 && minor == 6 && micro < 25))
+    return;
+
+  fs_2625 = 1;
+}
+
 void inline_size
 infy_init (EV_P)
 {
   if (fs_fd != -2)
     return;
 
-  /* kernels < 2.6.25 are borked
-   * http://www.ussg.indiana.edu/hypermail/linux/kernel/0711.3/1208.html
-   */
-  {
-    struct utsname buf;
-    int major, minor, micro;
+  fs_fd = -1;
 
-    fs_fd = -1;
-
-    if (uname (&buf))
-      return;
-
-    if (sscanf (buf.release, "%d.%d.%d", &major, &minor, &micro) != 3)
-      return;
-
-    if (major < 2
-        || (major == 2 && minor < 6)
-        || (major == 2 && minor == 6 && micro < 25))
-      return;
-  }
+  check_2625 (EV_A);
 
   fs_fd = inotify_init ();
 
@@ -2586,7 +2642,7 @@ infy_fork (EV_P)
           if (fs_fd >= 0)
             infy_add (EV_A_ w); /* re-add, no matter what */
           else
-            ev_timer_start (EV_A_ &w->timer);
+            ev_timer_again (EV_A_ &w->timer);
         }
     }
 }
@@ -2651,16 +2707,12 @@ ev_stat_start (EV_P_ ev_stat *w)
   if (expect_false (ev_is_active (w)))
     return;
 
-  /* since we use memcmp, we need to clear any padding data etc. */
-  memset (&w->prev, 0, sizeof (ev_statdata));
-  memset (&w->attr, 0, sizeof (ev_statdata));
-
   ev_stat_stat (EV_A_ w);
 
-  if (w->interval < MIN_STAT_INTERVAL)
-    w->interval = w->interval ? MIN_STAT_INTERVAL : DEF_STAT_INTERVAL;
+  if (w->interval < MIN_STAT_INTERVAL && w->interval)
+    w->interval = MIN_STAT_INTERVAL;
 
-  ev_timer_init (&w->timer, stat_timer_cb, w->interval, w->interval);
+  ev_timer_init (&w->timer, stat_timer_cb, 0., w->interval ? w->interval : DEF_STAT_INTERVAL);
   ev_set_priority (&w->timer, ev_priority (w));
 
 #if EV_USE_INOTIFY
@@ -2670,7 +2722,7 @@ ev_stat_start (EV_P_ ev_stat *w)
     infy_add (EV_A_ w);
   else
 #endif
-    ev_timer_start (EV_A_ &w->timer);
+    ev_timer_again (EV_A_ &w->timer);
 
   ev_start (EV_A_ (W)w, 1);
 
@@ -2855,11 +2907,16 @@ embed_fork_cb (EV_P_ ev_fork *fork_w, int revents)
 {
   ev_embed *w = (ev_embed *)(((char *)fork_w) - offsetof (ev_embed, fork));
 
+  ev_embed_stop (EV_A_ w);
+
   {
     struct ev_loop *loop = w->other;
 
     ev_loop_fork (EV_A);
+    ev_loop (EV_A_ EVLOOP_NONBLOCK);
   }
+
+  ev_embed_start (EV_A_ w);
 }
 
 #if 0
